/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is JSIRC Library
 *
 * The Initial Developer of the Original Code is New Dimensions Consulting,
 * Inc. Portions created by New Dimensions Consulting, Inc. are
 * Copyright (C) 1999 New Dimenstions Consulting, Inc. All
 * Rights Reserved.
 *
 * Contributor(s):
 *  Robert Ginda, rginda@ndcico.com, original author
 *  Peter Van der Beken, peter.vanderbeken@pandora.be, necko-only version
 *
 * depends on utils.js, and XPCOM/XPConnect in the JS environment
 *
 */

var jsenv = new Object();
jsenv.HAS_DOCUMENT = (typeof document == "object");
jsenv.HAS_NSPR_EVENTQ = jsenv.HAS_DOCUMENT;
jsenv.HAS_STREAM_PROVIDER = ("nsIStreamProvider" in Components.interfaces);


// HACK:
//   I am changing this to true because the dict protocol handler
//   doesn't work otherwise.  I supsect it should be true, anyway,
//   and that testing if 'document' is defined is not a good way
//   to determine if we have an event queue.
jsenv.HAS_NSPR_EVENTQ = true;

function toScriptableInputStream (i)
{
    var si = Components.classes["@mozilla.org/scriptableinputstream;1"];
    
    si = si.createInstance();
    si = si.QueryInterface(Components.interfaces.nsIScriptableInputStream);
    si.init(i);

    return si;
    
}

function CBSConnection ()
{
    var sockServiceClass =
        Components.classesByID["{c07e81e0-ef12-11d2-92b6-00105a1b0d64}"];
    
    if (!sockServiceClass)
        throw ("Couldn't get socket service class.");
    
    var sockService = sockServiceClass.getService();
    if (!sockService)
        throw ("Couldn't get socket service.");

    this._sockService = sockService.QueryInterface
        (Components.interfaces.nsISocketTransportService);

    this.wrappedJSObject = this;

}

CBSConnection.prototype.connect =
function bc_connect(host, port, bind, tcp_flag, observer)
{
    if (typeof tcp_flag == "undefined")
		tcp_flag = false;
    
    this.host = host.toLowerCase();
    this.port = port;
    this.bind = bind;
    this.tcp_flag = tcp_flag;

    if (jsenv.HAS_STREAM_PROVIDER)
    {
        this._transport = this._sockService.createTransport(host, port, null, 0, 0);

        if (!this._transport)
            throw "Error creating transport.";

        if (jsenv.HAS_NSPR_EVENTQ)
        {
            // we've got an event queue, so start up an async write.
            this._streamProvider = new StreamProvider(observer);
            this._write_req = this._transport.asyncWrite(this._streamProvider, this, 0, -1, 0);
        }
        else
        {   /* no nspr event queues in this environment, we can't use async calls,
             * so set up the streams. */
            this._outputStream = this._transport.openOutputStream(0, -1, 0);
            if (!this._outputStream)
                throw "Error getting output stream. (no async, stream provider)";

            this._inputStream = this._transport.openInputStream (0, -1, 0);
            if (!this._inputStream)
                throw "Error getting input stream. (no async, stream provider)";
        }
    }
    else
    {
        // use new necko interfaces (bug 111479)
        this._transport = this._sockService.createTransport(null, 0, host, port, null);
        if (!this._transport)
            throw "Error creating transport.";

        var openFlags;
        if (jsenv.HAS_NSPR_EVENTQ)
            openFlags = 0;
        else
            openFlags = Components.interfaces.nsITransport.OPEN_BLOCKING;

        this._outputStream = this._transport.openOutputStream(openFlags, 4096, 32);
        if (!this._outputStream)
            throw "Error getting output stream. (no stream provider)";

        this._inputStream = this._transport.openInputStream(openFlags, 0, 0);
        if (!this._inputStream)
            throw "Error getting input stream. (no stream provider)";
    }

    this.connectDate = new Date();
    this.isConnected = true;

    return this.isConnected;
  
}

CBSConnection.prototype.disconnect =
function bc_disconnect()
{    
    if ("_inputStream" in this && this._inputStream)
        this._inputStream.close();
    if ("_outputStream" in this && this._outputStream)
        this._outputStream.close();

    this.isConnected = false;
}

CBSConnection.prototype.sendData =
function bc_senddata(str)
{
    if (!this.isConnected)
        throw "Not Connected.";

    if (jsenv.HAS_NSPR_EVENTQ && jsenv.HAS_STREAM_PROVIDER)
        this.asyncWrite(str);
    else
        this.sendDataNow(str);
}
    
CBSConnection.prototype.readData =
function bc_readdata(timeout, count)
{
    if (!this.isConnected)
        throw "Not Connected.";

    var rv;

    try
    {
        rv = this._scriptableInputStream.read(count);
    }
    catch (ex)
    {
        //dd("*** Caught " + ex + " while reading.")
        this.isConnected = false;
        throw ex;
    }
    
    return rv;
}

CBSConnection.prototype.startAsyncRead =
function bc_saread (observer)
{
    if (jsenv.HAS_STREAM_PROVIDER)
    {
        this._transport.asyncRead(new StreamListener(observer), this, 0, -1, 0);
    }
    else
    {
        var cls = Components.classes["@mozilla.org/network/input-stream-pump;1"];
        var pump = cls.createInstance(Components.interfaces.nsIInputStreamPump);
        pump.init(this._inputStream, -1, -1, 0, 0, false);
        pump.asyncRead(new StreamListener(observer), this);
    }

}

CBSConnection.prototype.asyncWrite =
function bc_awrite (str)
{
    this._streamProvider.pendingData += str;
    if (this._streamProvider.isBlocked)
    {
        this._write_req.resume();
        this._streamProvider.isBlocked = false;
    }
}

CBSConnection.prototype.hasPendingWrite =
function bc_haspwrite ()
{
    if (jsenv.HAS_STREAM_PROVIDER)
        return (this._streamProvider.pendingData != "");
    else
        return false; // data already pushed to necko
}

CBSConnection.prototype.sendDataNow =
function bc_senddatanow(str)
{
    var rv = false;
    
    try
    {
        this._outputStream.write(str, str.length);
        rv = true;
    }
    catch (ex)
    {
        //dd ("*** Caught " + ex + " while sending.")
        this.isConnected = false;
        throw (ex);
    }
    
    return rv;
}

function _notimpl ()
{
    throw "Not Implemented.";
}

if (!jsenv.HAS_NSPR_EVENTQ)
{
    CBSConnection.prototype.startAsyncRead = _notimpl;
    CBSConnection.prototype.asyncWrite = _notimpl;
}
else if (jsenv.HAS_STREAM_PROVIDER)
{
    CBSConnection.prototype.sendDataNow = _notimpl;
}
else
{
    CBSConnection.prototype.asyncWrite = _notimpl;
}

delete _notimpl;
    
function StreamProvider(observer)
{
    this._observer = observer;
}

StreamProvider.prototype.pendingData = "";
StreamProvider.prototype.isBlocked = true;

StreamProvider.prototype.onDataWritable =
function sp_datawrite (request, ctxt, ostream, offset, count)
{
    if ("isClosed" in this && this.isClosed)
        throw Components.results.NS_BASE_STREAM_CLOSED;

    if (!this.pendingData)
    {
        this.isBlocked = true;

        /* this is here to support pre-XPCDOM builds (0.9.0 era), which
         * don't have this result code mapped. */
        if (!Components.results.NS_BASE_STREAM_WOULD_BLOCK)
            throw 2152136711;
        
        throw Components.results.NS_BASE_STREAM_WOULD_BLOCK;
    }
    
    var len = ostream.write (this.pendingData, this.pendingData.length);
    this.pendingData = this.pendingData.substr (len);
}

StreamProvider.prototype.onStartRequest =
function sp_startreq (request, ctxt)
{
}


StreamProvider.prototype.onStopRequest =
function sp_stopreq (request, ctxt, status)
{
    if (this._observer)
        this._observer.onStreamClose(status);
}

function StreamListener(observer)
{
    this._observer = observer;
}

StreamListener.prototype.onStartRequest =
function sl_startreq (request, ctxt)
{
}

StreamListener.prototype.onStopRequest =
function sl_stopreq (request, ctxt, status)
{
    if (this._observer)
        this._observer.onStreamClose(status);
}

StreamListener.prototype.onDataAvailable =
function sl_dataavail (request, ctxt, inStr, sourceOffset, count)
{
    ctxt = ctxt.wrappedJSObject;
    if (!ctxt)
    {
        //dd ("*** Can't get wrappedJSObject from ctxt in " +
        //    "StreamListener.onDataAvailable ***");
        return;
    }

    if (!("_scriptableInputStream" in ctxt))
        ctxt._scriptableInputStream = toScriptableInputStream(inStr);

    if (this._observer)
        this._observer.onStreamDataAvailable(request, inStr, sourceOffset, count);
}

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// This code in this file converts multi-byte encodings from UTF-8 into HTML
// character entities of the form &#####;, where ##### is the number
// that the multi-byte sequence encodes.
//
// This is a work-around for a problem that I haven't been able to
// find the right solution to.  The problem is that the Dict extension
// is unable to tell Mozilla/Firefox that the content it is displaying
// is UTF-8.  Mozilla/Firefox then mis-interprets it as an 8-bit coding
// from the ISO-8859-1 (Latin-1_ character set, causing garbage to be 
// displayed for words with non-ASCII characters.


function dictUtf8ToCharacterEntitiesInternal(utf8String) {

    var encodedString = '';
    var octetsRemainingInCharacterCode = 0;
    var characterCode = 0;

    for (var i = 0; i < utf8String.length; i++) {

        var inputCharCode = utf8String.charCodeAt(i);

        if ((inputCharCode & 0x80) == 0x00) {
            // this can be encoded with ASCII-7
            encodedString += utf8String[i];

            // reset the multi-octet stream accumulator.
            // if it was not zero, then the input was bad.
            octetsRemainingInCharacterCode = 0;
        }
        else if ((inputCharCode & 0xC0) == 0x80) {
            // 10xxxxxx

            // this is a continuation of a multi-octet sequence
            characterCode = (characterCode << 6) | (inputCharCode & 0x3F);

            octetsRemainingInCharacterCode--;
            if (octetsRemainingInCharacterCode == 0) {
                // this is the final octet in the mutli-octet code
                encodedString += "&#" + characterCode + ";";
            }
            else if (octetsRemainingInCharacterCode < 0) {
                // bad input
                octetsRemainingInCharacterCode = 0;
            }
        }
        else if ((inputCharCode & 0xE0) == 0xC0) {
            // 110xxxxx 10xxxxxx
            characterCode = inputCharCode & 0x1F;
            octetsRemainingInCharacterCode = 1;
        }
        else if ((inputCharCode & 0xF0) == 0xE0) {
            // 1110xxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x0F;
            octetsRemainingInCharacterCode = 2;
        }
        else if ((inputCharCode & 0xF8) == 0xF0) {
            // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x07;
            octetsRemainingInCharacterCode = 3;
        }
        else if ((inputCharCode & 0xFC) == 0xF8) {
            // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x03;
            octetsRemainingInCharacterCode = 4;
        }
        else if ((inputCharCode & 0xFE) == 0xFC) {
            // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x01;
            octetsRemainingInCharacterCode = 5;
        }
        else {
            // Illegal UTF-8 code.  Drop it.
        }
    }

    return encodedString;
}

//
// dictUtf8ToCharacterEntities() converts an octet stream that is 
// encoded in UTF-8 into an ASCII-7 encoding of the same 
// stream such that all character > 127 are encoded using
// HTML character entities &#xxx;.
//
// This is necessary when using international DICT servers,
// because I cannot figure out another way to tell Mozilla
// to interpret the string as UTF-8.
//
//   UCS-4 range (hex.)    UTF-8 octet sequence (binary)
//   0000 0000-0000 007F   0xxxxxxx
//   0000 0080-0000 07FF   110xxxxx 10xxxxxx
//   0000 0800-0000 FFFF   1110xxxx 10xxxxxx 10xxxxxx
//   0001 0000-001F FFFF   11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0020 0000-03FF FFFF   111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0400 0000-7FFF FFFF   1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//
function dictUtf8ToCharacterEntities(utf8String) {

    // Before we do any convertion, let's see if there are
    // any character codes outside of the ASCII-7 range.
    // If not, then the string doesn't need to be converted.

    if (!utf8String.match(/[\x80-\xFF]/)) {
        return utf8String;
    }

    // Note that we could just call utf8ToCharacterEntitiesInternal()
    // on the entire string, but this routine is very slow.
    // Instead, it is faster to use a regular expression
    // to separate the octet-sequences that must be replaced
    // while preserving long runs of characters that don't to escaping.

    // This makes the routine run in O(n^2) of the characters that
    // must be replaced, instead of O(n^2) of all characters.
    // When the text is mostly ASCII-7, this is a significant savings.

    var encodedString = '';
    
    var matchExp = /([\x00-\x7F]*)([\x80-\xFF]+)([\x00-\x7F]*)/g;
    while ((result = matchExp.exec(utf8String)) != null) {

        var singleOctetCodesBefore = result[1];
        var multiOctetCodes        = result[2];
        var singleOctetCodesAfter  = result[3];
        
        encodedString += 
            singleOctetCodesBefore + 
            dictUtf8ToCharacterEntitiesInternal(multiOctetCodes) +
            singleOctetCodesAfter;
    }

    return encodedString;
}

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// put constants into a namespace so we don't pollute the global namespace
function DictState () {
    
    // class contants
    this.Start                = 0;
    this.WantClientAck        = 1;
    this.WantDefinition       = 2;
    this.WantDefinitionsDone  = 3;
    this.WantStrategiesInfo   = 4;
    this.WantStrategies       = 5;
    this.WantMatchesInfo      = 6;
    this.WantMatches          = 7;
    this.Done                 = 8;
    this.Error                = -1;
}

const DICTSTATE = new DictState();

function Observer(listener, connection, word, database) {

    this._listener   = listener;
    this._connection = connection;
    this._word       = word;
    this._database   = database;

    this._bufferedData = '';
    this._state = DICTSTATE.Start;

    this._definitions  = {};

    this._createDate = new Date();
}


// Reports that there was a protocol error.
// This should only happen if the server doesn't obey the protocol
// (or if there are bugs in the client state machine)
Observer.prototype.reportStateMachineError = function ()
{
    this._state = DICTSTATE.Error;

    throw 
        'Unknown problem encountered when defining "' + this._word + '".\n' +
        'If problem presists, please file a bug report on the Dict project.\n';
}

// Reports that the requested database does not exist on the server
Observer.prototype.reportUnknownDatabase = function ()
{
    this._state = DICTSTATE.Error;

    throw 
        this._connection.host + ' does not have a database called "' + this._database + '".\n' +
        'Please specify a different database.\n';
}

// Reports that the server does not have any databases installed
Observer.prototype.reportNoDatabases = function ()
{
    this._state = DICTSTATE.Error;

    throw 
        this._connection.host + ' does not have any databases installed.\n' +
        'Please use a different server.\n';
}

// Reports that the client received an access denied error.
Observer.prototype.reportAccessDenied = function ()
{
    this._state = DICTSTATE.Error;

    throw 
        this._connection.host + ' requires user authentication, but ' + 
        'the Dict plugin does not support authentication.\n' + 
        'If this feature is important to you, please let us know by filing a bug on the Dict project.';
}

// Reports that the client received an access denied error.
Observer.prototype.reportTemporarilyUnavailable= function ()
{
    this._state = DICTSTATE.Error;

    throw 
        this._connection.host + ' is temporarily unavailable, but expects to be available soon.\n' +
        'Please wait a short while and retry your request.';
}


// Reports that the client received an access denied error.
Observer.prototype.reportServerShuttingDown = function ()
{
    this._state = DICTSTATE.Error;

    throw 
        this._connection.host + ' is being shut down and will not be able to define\n' + 
        'words in the near-future.\n' +
        'Please change your preferences to use a different server.';
}

// Pops a DICT response line from the _bufferedInput member variable.
// On success, returns true and sets this._responseLine, this._responseCode, 
// and this._responseLineBody.
// Return false, if there wasn't a complete response line.
Observer.prototype.popResponseLine = function ()
{
    var responseLinePattern = /^(\d+)\s*(.*)\n/;
    
    var result = this._bufferedData.match(responseLinePattern);
    if (result == null) {
        return false;
    }

    this._responseLine     = result[0];
    this._responseCode     = result[1];
    this._responseLineBody = result[2];
    
    // remove the response line from the buffered input
    var lineEndIndex   = this._bufferedData.indexOf('\n');
    this._bufferedData = this._bufferedData.slice(lineEndIndex + 1);

    // Handle error codes that should result in immediately
    // reporting an error to the user regardless of the current state.
    if (this._responseCode == 530 ||
        this._responseCode == 531 ||
        this._responseCode == 532) {
        // access denied
        this.reportAccessDenied();
    }
    else if (this._responseCode == 420) {
        // server is temporarily unavailable
        this.reportTemporarilyUnavailable();
    }
    else if (this._responseCode == 421) {
        // server is temporarily unavailable
        this.reportServerShuttingDown();
    }
    else if (this._responseCode == 550) {
        // invalid database
        this.reportUnknownDatabase();
    }
    else if (this._responseCode == 554) {
        // no databases present
        this.reportNoDatabases();
    }

    return true;
}

// Performs a graceful shutdown
Observer.prototype.gracefulShutdown = function ()
{
    this._connection.sendData('QUIT\r\n');
    this._bufferedData = '';
    this._state = DICTSTATE.Done;
    
    this._connection.disconnect();
}


// Parses and returns all buffered input up to (but not including) 
// the sequence CRLF . CRLF.  The multiline input is then removed 
// from the buffered input.
// Returns null if the multiline input is not complete.
//
// The DICT protocol uses multiline input of this form to return 
// strategies, matches, and definitions. 
Observer.prototype.popMultilineInput = function () 
{
    // NOTE: We should like for \r\n.\r\n, but all "\r" are removed
    //       from the input stream to make regexp parsing easier.
    const sentinal = "\n.\n";

    // Check to see if we have the entire list
    var endIndex = this._bufferedData.indexOf(sentinal);
    if (endIndex == -1) {
        return null;  // the multiline input isn't complete
    }
    
    // pop off and return the multiline input
    var multilineData = this._bufferedData.slice(0, endIndex);
    
    // remove the multiline input and sentinal sequence from the buffered input
    this._bufferedData = this._bufferedData.slice(endIndex + sentinal.length);

    // According to RFC 2229, Section 2.4.3, if a line of text begins 
    // with two periods, the client must treat this as a single, 
    // literal period.
    multilineData = multilineData.replace(/^\.\./gm, '.');

    return multilineData;
}

Observer.prototype.onStreamDataAvailable = function (
    request, 
    inStream, 
    offset, 
    count
) 
{
    if (this._state == DICTSTATE.Done ||
        this._state == DICTSTATE.Error) {
        // sometimes if we perform a graceful shutdown, this handler
        // gets called again.  Any operation will throw an exception,
        // so it's best to just ignore the call.
        return;
    }

    try {
        // read the data from the socket into our buffer
        var newData = this._connection.readData(0, count);

        // strip out carriage returns, since they complicate regexp usage
        this._bufferedData += newData.replace(/\r/g, '');

        // do {
        //   {handle data}
        // } while( {there's enough data to handle} );
        do {
            var previousLength = this._bufferedData.length;

            var matchExp;
            var word;

            switch (this._state) {

            case DICTSTATE.Start:
                // get the 220 "hello" response that is sent when we connect

                if (this.popResponseLine()) {
                    
                    if (this._responseCode == 220) {
                        // 220 <text> <capabilities> <msg-id> CRLF
                        // (this is the session init sequence)
                    
                        this._state = DICTSTATE.WantClientAck;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;


            case DICTSTATE.WantClientAck:
                // get the 250 ok resonse for our "CLIENT" command

                if (this.popResponseLine()) {
                    if (this._responseCode == 250) {
                        this._state = DICTSTATE.WantDefinitionsDone;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantDefinitionsDone:

                if (this.popResponseLine()) {
                    if (this._responseCode == 150) {
                        // 150: n definitions returned
                    }
                    else if (this._responseCode == 151) {
                        // 151: <word> <database> <name>
                        // The definition follows this line
                        this._state = DICTSTATE.WantDefinition;
                    }
                    else if (this._responseCode == 250) {

                        // We have all definitions
                        this._listener.onHaveDefinitions(
                            this._word,
                            this._definitions);

                        this.gracefulShutdown();
                    }
                    else if (this._responseCode == 550) {
                        // 550 invalid database

                        // TODO: Handle this gracefully
                        this.reportStateMachineError();
                    }
                    else if (this._responseCode == 552) {
                        // 552 no match

                        // Present the user with a nearest match 
                        
                        // The "*" database means search through all databases.
                        //
                        // The "." strategy means "use whatever is best for
                        // for interactive spell-checking", according to 
                        // RFC 2229, section 3.3.1.
                        //
                        this._connection.sendData('MATCH * . "' + this._word + '"\r\n');
                        this._state = DICTSTATE.WantMatchesInfo;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantDefinition:
                
                // 151 <word> <database> CRLF
                // <definition line 1>   CRLF
                // <definition line 2>   CRLF
                // . . . 
                // <definition line n>   CRLF
                // . CRLF
                
                // determine if we have the entire definition
                var definitionText = this.popMultilineInput();
                if (definitionText == null) {
                    break; // wait for the rest of the definition
                }

                var definition = new Object;

                // 151 <word> <database> CRLF
                var result = this._responseLineBody.match(/^(\S+)\s+\S+\s+(.*)$/);
                if (result != null) {
                    definition.word   = result[1];
                    definition.source = result[2];
                }

                // Store the definition
                definition.text = definitionText;

                // Add this definition to the set of definitions.
                // this._definitions maps the source name to an array of 
                // definition objects.
                if (this._definitions[definition.source] != null) {
                    this._definitions[definition.source].push(definition);
                }
                else {
                    this._definitions[definition.source] = new Array ();
                    this._definitions[definition.source].push(definition);
                }

                this._state = DICTSTATE.WantDefinitionsDone;
                break;

            case DICTSTATE.WantStrategiesInfo:

                // DEAD CODE: 
                //   This state is not reachable, since the code never
                //   sends the "SHOW STRAT" command.  However,
                //   this state code is correct, so I have kept it intact.

                if (this.popResponseLine()) {
                    if (this._responseCode == 111) {
                        // 111 n stragegies available
                        this._state = DICTSTATE.WantStrategies;
                    }
                    else if (this._responseCode == 555) {
                        // 555 No strategies available
                        this.gracefulShutdown();
                        this._listener.onNoDefinitionsFound(this._word);
                    }
                    else {
                        // unexpected response
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantStrategies:

                // DEAD CODE: 
                //   This state is not reachable, since the code never
                //   sends the "SHOW STRAT" command.  However,
                //   the state code is correct, so I have kept it intact.

                // <strategy 1> <description 1> CRLF
                // <strategy 2> <description 2> CRLF
                // . . . 
                // <strategy n> <description n> CRLF
                // . CRLF

                // determine if we have the entire strategy list
                var strategies = this.popMultilineInput();
                if (strategies == null) {
                    break; // wait for the rest of the strategies
                }


                // search for the "Levenshtein distance one"
                var haveLevStrategy = false;
                // <strategy> <description> CRLF
                matchExp = /^(\S+)\s+(.*)$/gm;
                while ((result = matchExp.exec(strategies)) != null) {
                    var strategy    = result[1];
                    var description = result[2];

                    if (strategy == 'lev') {
                        haveLevStrategy = true;
                    }
                }

                if (haveLevStrategy) {
                    // Use the "lev" strategy to find nearest matches
                    this._connection.sendData('MATCH * lev "' + this._word + '"\r\n');
                    this._state = DICTSTATE.WantMatchesInfo;
                }
                else {
                    // No good matching strategies
                    this.gracefulShutdown();
                    
                    this._listener.onNoDefinitionsFound(this._word);
                }

                break;

            case DICTSTATE.WantMatchesInfo:

                if (this.popResponseLine()) {
                    if (this._responseCode == 250) {
                        // 250 ok
                        // This is from the MATCH request.  Ignore it.
                    }
                    else if (this._responseCode == 152) {
                        // 152 n matches found
                        this._state = DICTSTATE.WantMatches;
                    }
                    else if (this._responseCode == 552) {
                        // 552 no match
                        this.gracefulShutdown();
                        this._listener.onNoDefinitionsFound(this._word);
                    }
                    else {
                        // unexpected response
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantMatches:

                // <database 1> <word 1> CRLF
                // <database 2> <word 2> CRLF
                // . . . 
                // <database n> <word n> CRLF
                // . CRLF
                
                // determine if we have the entire strategy list
                var matches = this.popMultilineInput();
                if (matches == null) {
                    break; // wait for the rest of the matches
                }

                // <database> <word> CRLF
                var words    = new Object;
                matchExp = /^(\S+)\s+(.*)$/gm;
                while ((result = matchExp.exec(matches)) != null) {
                    var database    = result[1];
                    word            = result[2];

                    var normalizedWord = word.toLowerCase();

                    words[normalizedWord] = true;
                }

                // REVISIT: There has got to be a more direct way
                //   of counting the number of elements in an associative
                //   array.
                var totalMatches = 0;
                for (word in words) {
                    totalMatches++;
                }

                if (totalMatches == 0) {
                    // There are no matches.

                    // This is probably a bug in either the server or the
                    // client, since we should not have transitioned to
                    // the WantMatches state if there were no matches.
                    this.gracefulShutdown();

                    this._listener.onNoDefinitionsFound(this._word);
                }
                else if (totalMatches == 1) {
                    // There is exactly one match.
                    // Rather than present just one match, we can save the
                    // user some trouble and just request the definition
                    // of the match.

                    for (word in words) {
                        // Request a defintion for the only match
                        this._connection.sendData('DEFINE * ' + word + '\r\n');
                        this._word  = word;
                        this._state = DICTSTATE.WantClientAck;
                    }
                }
                else {
                    // There is more than one close match
                    this._listener.onHaveClosestMatches(this._word, words);

                    this.gracefulShutdown();
                }

                break;

            case DICTSTATE.Done:
                break;

            default:
                reportStateMachineError();
                break;
            }
            
        } while (previousLength != this._bufferedData.length);
    }
    catch (ex) {
        this._connection.disconnect();
        this._listener.onError(ex);
    }
}

Observer.prototype.onStreamClose = function (
    status
) 
{
    var closeDate = new Date();
    var delta = closeDate.getTime() - this._createDate.getTime();
    //alert('query took ' + delta + ' ms');

    if (this._state != DICTSTATE.Done &&
        this._state != DICTSTATE.Error) {
        alert(
            'Unable to connect to ' + this._connection.host + '\n' + 
            'If the problem persists, try using a different server.');
    }
}


function DictClient() {
}

DictClient.prototype = 
{

    // defineWord() is the main entry point to the DICT client.
    // This function contacts the DICT server and asks for the
    // definition of a given word.  It is asynchornous and the
    // definitions (or errors) are returned by callbacks on
    // the "listener" parameter.
    //
    // host - 
    //     The host running the DICT server
    //
    // port - 
    //     The port on which the DICT server is listening.
    //     This is usually 2628.
    //
    // word - 
    //     The word to define
    // 
    // listener - 
    //     A special "listener" object which handles the server's response.
    //     This object must define the following functions:
    //
    //       onHaveDefinitions: function (word, definitions)
    //          Called when DictClient has the definitions.
    //
    //          word - 
    //              The word that was actually defined.
    //              This may be different from the requested word, if
    //              the exact word was not found and only one match 
    //              existed.
    //
    //          definitions - 
    //              An associated array that maps the name of the definition's source
    //              (the dictionary) to an array of definitions.
    //              Each definition is an object with one field "text"
    //              which is a UTF-8 encoded definition of the word.
    //
    //              The following code shows how to iterate over all definitions:
    //
    //                for (var source in definitions) {
    // 
    //                  var i;
    //                  for (i = 0; i < definitions[source].length; i++) {
    //
    //                    var def = definitions[source][i].text;
    //                    <do something with def>
    //                  }
    //                }
    //
    //       onHaveClosestMatches: function (word, matches)
    //          Called when DictClient has a list of close matches.
    //
    //          word    - 
    //              the word for which there are close matches.
    //
    //          matches - 
    //              an array of strings--each of which is a close match
    //              for word that has a definition.
    //
    //       onNoDefinitionsFound : function (word) 
    //          Called when the DictClient could not find any definitions.
    //
    //          word - the word passed into defineWord().
    //
    //       onError : function (error) 
    //          Called when the DictClient could not define the word.
    //
    //          error - plain-text explaination of what went wrong.
    //
    // database - [optional]
    //    The database to define the word.
    //    If this parameter is not given, '*' will be used.
    //
    defineWord : function(host, port, word, listener, database)
    {
        // if no database was specified, default to '*' (all databases)
        if (database == null) {
            database = '*';
        }

        var connection = new CBSConnection();
    
        connection.connect(host, port, (void 0), true, null);

        var observer = new Observer(listener, connection, word, database);

        // To save time, pipe-line the CLIENT and DEFINE commands,
        // rather than wait for the "220 ok" handshake.
        // The server is required to support command pipe-lining.
        // This makes a measurable difference there is high latency.
        connection.sendData('CLIENT Mozilla Dict plugin\r\n');
        connection.sendData('DEFINE ' + database + ' "' + word + '"\r\n');
        connection.startAsyncRead(observer);
    }
}

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// HACK:
//   I want to be able to call functions that are defined within my JAR file,
//   but I can't figure out how to do this within JavaScript.
//   Instead, the Build system will concatenate the files that I want to call,
//   along with this file, and create dict-services.js.
//   This is ugly, but it is more maintainable than copying that code here.
//

//
// This file contains a protocol handler for supplying a channel to the browser
// when an dict:// URL is used.
//

// classes/services defined in this file.
const DICTPROT_HANDLER_CONTRACTID = "@mozilla.org/network/protocol;1?name=dict";
const DICTPROT_HANDLER_CID        = Components.ID("{f39a1786-c513-4def-8833-d158ff9cd258}");

// components used in this file 
const STANDARDURL_CONTRACTID = "@mozilla.org/network/standard-url;1";
const IOSERVICE_CONTRACTID   = "@mozilla.org/network/io-service;1";
const STRINGINPUT_CONTRACTID = "@mozilla.org/io/string-input-stream;1";

// interfaces used in this file
const nsIProtocolHandler   = Components.interfaces.nsIProtocolHandler;
const nsIURI               = Components.interfaces.nsIURI;
const nsIChannel           = Components.interfaces.nsIChannel;
const nsIRequest           = Components.interfaces.nsIRequest;
const nsIIOService         = Components.interfaces.nsIIOService;
const nsISupports          = Components.interfaces.nsISupports;
const nsIInputStream       = Components.interfaces.nsIInputStream;
const nsIStringInputStream = Components.interfaces.nsIStringInputStream;


// DictListener class, which is passed in as the "listener" parameter
// to DictClient::defineWord().
function DictListener (channel, definitionIndex) {
    this._channel         = channel;
    this._definitionIndex = definitionIndex;
}

DictListener.prototype = 
{
    // wraps HTML inside HTML/BODY tag and adds a title
    createHtmlDocument : function (title, body)
    {
        var htmlDocument;
    
        htmlDocument = 
            '<HTML>\n'    +
            '  <HEAD>\n'  + 
            '      <TITLE>' + dictUtf8ToCharacterEntities(title) + '</TITLE>\n' +
            '  </HEAD>\n' +
            '  <BODY>\n'  +
            dictUtf8ToCharacterEntities(body) +
            '  </BODY>\n' +
            '</HTML>\n';
        

        return htmlDocument;
    },

    createHtmlDefinition : function (definition)
    {
        var html = definition;
        
        // normalize all definitions so that they end in a single newline
        html = html.replace(/\s*$/, '\n');

        // Escape HTML entities
        html = html.replace(/&/g, '&amp;');
        html = html.replace(/</g, '&lt;');

        // replace newlines with <BR> and whitespace with
        // non-breaking spaces to preserve formatting.
        // This is better than using <PRE> because it
        // allows us to use other mark-up tags.
        html = html.replace(/\n/g, '<BR>');
        html = html.replace(/ /g, '&nbsp;');

        // Words that are cross-referenced by WordNet show up 
        // in braces. e.g., "{word}" and "{word(P)}"
        html = html.replace(/{(.*?)(\(.*\))?}/g, '$1');

        // add a newline after each <BR> to make the 
        // HTML source more readable.
        html = html.replace(/<BR>/g, '<BR>\n');

        return html;
    },

  
    //
    // The next three functions are the DictListener interface
    //

    // Called when the DictClient has definitions
    onHaveDefinitions : function (word, definitions)
    {
        // Format the definitions into HTML
        var htmlBody = '<DIV class="dict-page">\n';

        // count the number of sources -- used for formatting
        var totalSources = 0;
        for (source in definitions) {
            totalSources++;
        }

        var sourceIndex = 0;
        for (source in definitions) {

            htmlBody += '<B><DIV class="dict-database">';
            htmlBody += source;
            htmlBody += '</DIV></B>\n';
            
            htmlBody += '<CODE><SPAN class="dict-definition">\n';

            var i;
            for (i = 0; i < definitions[source].length; i++) {

                if (this._definitionIndex) {
                    // the caller is only requesting one definition

                    var thisIndex = i + 1; // Dict definitions are 1-indexed
                    if (this._definitionIndex == thisIndex) {
                        htmlBody += this.createHtmlDefinition(definitions[source][i].text);
                    }
                } 
                else {
                    // the caller is requesting all definitions
                    var deftext = this.createHtmlDefinition(definitions[source][i].text);

                    // append an extra newline after everything but the
                    // the final definition from the final source.
                    // (Otherwise there'd be a strange-looking blank space
                    // at the bottom of the page.)
                    if (sourceIndex + 1 != totalSources ||
                        i + 1           != definitions[source].length) {
                        deftext += '<BR>\n';
                    }

                    htmlBody += deftext;
                }
            }

            htmlBody += '</SPAN></CODE>\n';
            sourceIndex++;
        }

        htmlBody += '</DIV>\n';

        // create an HTML document
        var htmlDocument = this.createHtmlDocument(word, htmlBody);
        
        // give the HTML document to the channel
        this._channel.Write(htmlDocument);
    },

    // Called when the DictClient has closest matches
    onHaveClosestMatches : function (word, matches)
    {
        // There is more than one closest match
        // Display them all and let the user decide which to define
        var htmlBody = 
            '<DIV class="dict-page" style="width:450">\n' + 
                '<B><DIV class="dict-heading">Closest Matches</DIV></B><BR>\n' + 
                '<DIV class="dict-text">\n' + 
                    'No definitions for "' + word + '" were found, ' +
                    'but definitions for similar-looking words were found.  ' +
                    'To define one of these close matches, highlight ' +
                    'the word below, right-click, and select "Define" ' +
                    'from the pop-up menu.' +
                 '</DIV>\n';

        // add a spacer between the text and list of words
        htmlBody += '<HR class="dict-horizontal-rule">\n';

        for (match in matches) {
            htmlBody += match + '<BR>\n';
        }

        htmlBody += '</DIV>\n';
        
        var htmlDocument = this.createHtmlDocument(
            word + ' - Closest Matches',
            htmlBody);
        
        // give the HTML document to the channel
        this._channel.Write(htmlDocument);
    },

    // called when the DictClient could not find any definitions
    onNoDefinitionsFound : function (word) 
    {
        // create an HTML document
        var htmlDocument = this.createHtmlDocument(
            'No Definitions Found', 
            'No definitions were found for "' + word + '".\n');

        // give the HTML document to the channel
        this._channel.Write(htmlDocument);
    },

    // called when the DictClient encounters some unrecoverable error
    onError : function (error)
    {
        // create an HTML document
        var htmlDocument = this.createHtmlDocument(
            'Error', 
            '<PRE>' + error + '</PRE>');
        
        // give the HTML document to the channel
        this._channel.Write(htmlDocument);
    }
}


// Channel used by the DictProtocol
function DictChannel (input_uri)
{
    this.URI         = input_uri;
    this.originalURI = input_uri;
}

DictChannel.prototype = 
{
    // nsISupports
    QueryInterface : function (iid)
    {
        if (!iid.equals(nsIChannel) && 
            !iid.equals(nsIRequest) &&
            !iid.equals(nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }

        return this;
    },


    // called by DictListener when it has received data from DictClient
    Write : function (text)
    {
        var stringInputStreamClass = Components.classes[STRINGINPUT_CONTRACTID];

        var inputStream = stringInputStreamClass.getService(nsIInputStream);
        var stringInputStream = inputStream.QueryInterface(nsIStringInputStream);

        stringInputStream.setData(text, text.length);

        this._listener.onStartRequest(this, this._listenerContext);
        this._listener.onDataAvailable(this, this._listenerContext, inputStream, 0, text.length);
        //this._listener.onStopRequest(this, this._listenerContext);
    },

    // nsIChannel
    loadAttributes        : null,
    contentType           : "text/html",
    contentLength         : 0,
    owner                 : null,
    loadGroup             : null,
    notificationCallbacks : null,
    securityInfo          : null,

    open : function () 
    {
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    },


    asyncOpen : function (observer, context)
    {
        this._listener        = observer;
        this._listenerContext = context;

        observer.onStartRequest(this, context);

        var channel = this;

        //   dict://<user>;<auth>@<host>:<port>/d:<word>:<database>:<n>
        //   dict://<user>;<auth>@<host>:<port>/m:<word>:<database>:<strat>:<n>

        var host = channel.URI.host;

        var port = channel.URI.port;
        if (port == -1) {
            // if no port is specified, use the default port
            port = 2628; 
        }

        var path = channel.URI.path;

        const defineUriPattern = /^\/d:([^:]+):?([^:]+)?:?([^:]+)?$/;
        const matchUriPattern  = /^\/m:([^:]+):?([^:]+)?:?([^:]+)?:?([^:]+)?$/;

        var result;
        var word;
        var database;
        var index;
        var strategy;

        result = path.match(defineUriPattern);
        if (result != null) {
            // this is a "define" request.
            word     = result[1];
            database = result[2];
            index    = result[3];

            // Accoring to RFC 2229, Section 5:
            // If database is omitted, "!" SHOULD be used.
            // If index is omitted, then all definitions should be shown.

            if (database == null) {
                database = '!';
            }

            var listener = new DictListener(this, index);
            var client   = new DictClient();
            client.defineWord(host, port, word, listener, database);
        }

        result = path.match(matchUriPattern);
        if (result != null) {
            // this is a "match" request.

            word     = result[1];
            database = result[2];
            strategy = result[3];
            index    = result[4];

            // Accoring to RFC 2229, Section 5:
            // If database is omitted, "!" SHOULD be used.
            // If strategy is omitted, "." SHOULD be used.
            // If index is omitted, then all definitions should be shown.

            if (database == null) {
                databse = '!';
            }
            if (strategy == null) {
                strategy = '.';
            }

            // TODO: support matching
        }

    },

    asyncRead : function (listener, ctxt)
    {
        return listener.onStartRequest (this, ctxt);
    },


    // nsIRequest
    isPending : function ()
    {
        return true;
    },

    status : Components.results.NS_OK,

    cancel : function (status)
    {
        this.status = status;
    },

    suspend : function () 
    {
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    },

    resume : function () 
    {
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    }
};


// dict protocol handler 
function DictProtocol()
{
}

DictProtocol.prototype =
{
    QueryInterface: function(iid)
    {
        if (!iid.equals(nsIProtocolHandler) &&
            !iid.equals(nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    },

    scheme: "dict",

    defaultPort: 2628,

    protocolFlags: nsIProtocolHandler.URI_NORELATIVE |
                   nsIProtocolHandler.URI_NOAUTH,
  
    allowPort: function(port, scheme)
    {
        return false;
    },

    newURI: function(spec, charset, baseURI)
    {
        var uri = Components.classes[STANDARDURL_CONTRACTID].createInstance(nsIURI);
        uri.spec = spec;

        return uri;
    },

    newChannel: function(input_uri)
    {
        ios = Components.classes[IOSERVICE_CONTRACTID].getService(nsIIOService);
        if (!ios.allowPort(input_uri.port, input_uri.scheme)) {
            throw Components.results.NS_ERROR_FAILURE;
        }

        return new DictChannel(input_uri);
    }
}

// singleton class factory object for DictProtocol objects
var DictProtocolFactory = new Object();

DictProtocolFactory.createInstance = function (outer, iid)
{
    if (outer != null) {
        throw Components.results.NS_ERROR_NO_AGGREGATION;
    }

    if (!iid.equals(nsIProtocolHandler) && 
        !iid.equals(nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }

    return new DictProtocol();
}


// Dict Module
var DictModule = new Object();

DictModule.registerSelf = function (compMgr, fileSpec, location, type)
{
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);

    compMgr.registerFactoryLocation(
        DICTPROT_HANDLER_CID,
        "DICT Protocol handler",
        DICTPROT_HANDLER_CONTRACTID, 
        fileSpec, 
        location,
        type);
}

DictModule.unregisterSelf = function(compMgr, fileSpec, location)
{
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);

    compMgr.unregisterFactoryLocation(
        DICTPROT_HANDLER_CID,
        fileSpec);
}

DictModule.getClassObject = function (compMgr, cid, iid)
{
    if (cid.equals(DICTPROT_HANDLER_CID)) {
        return DictProtocolFactory;
    }

    if (cid.equals(DICTCNT_HANDLER_CID)) {
        return DictContentHandlerFactory;
    }
    
    if (iid.equals(Components.interfaces.nsIFactory)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }

    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
}

DictModule.canUnload = function(compMgr)
{
    return true;
}

// entrypoint
function NSGetModule(compMgr, fileSpec)
{
    return DictModule;
}

